//-------------------------------------------------------------------
//-------------------------------------------------------------------
//
// Simple Spring Mass System
// -- Matrix lib
//
// Primary Author: James F. O'Brien (obrienj@cc.gatech.edu)
//
// (C) Copyright James F. O'Brien, 1995
// (C) Copyright Georgia Institute of Technology, 1995
//-------------------------------------------------------------------
//-------------------------------------------------------------------
//
// RCS Revision History
//
// $Log: smMatrix.C,v $
// Revision 3.2  2003/10/13 20:23:57  job
// Finished port to Linux (Redhat 7.2).
// Moderate amount of testing, things appear to work.
//
// Revision 3.1  2003/10/11 04:08:42  job
// Rough port to Linux
//
// Revision 3.0  1999/03/11  22:29:14  obrienj
// Move from O32 ABI to n32
//
// Revision 2.7  1998/12/20  02:01:25  obrienj
// Eigen System routines
//
// Revision 2.6  1998/06/10  04:19:58  obrienj
// Added SmMatrix2x2 and moved rotation stuff into its own file.
//
// Revision 2.5  1998/02/19  20:57:51  obrienj
// Fixed bug in SmQuaternionToMoment.
// Minor restructuring.
//
// Revision 2.4  1997/10/25  22:44:07  obrienj
// Fixed BUG from Graphics Gems Matrix to Moment routine. (Grrrr!)
// Check for singularities in Quat to Moment and Moment to Quat.
//
// Revision 2.3  1997/10/08  22:52:04  obrienj
// Extended moment and quaternion support.
//
// Revision 2.2  1997/10/03  01:07:02  obrienj
// Added:  SmEulerAngToMatrixXYZ
//         SmMatrixToEulerAngXYZ
//         SmQuaternionToMatrix
//         SmMatrixToQuaternion
//
// Revision 2.1  1997/01/17  09:18:25  obrienj
// Made SmPosMatrixToMoment a bit more robust.
//
// Revision 1.2  1996/09/10  07:23:21  obrienj
// Fixed bug in operaror>> for SmMatrix3x3
// Minor change in format for SmMatrix3x3
// Added mat *= double operator
// Added mat * double operator
//
// Revision 1.1  1996/09/10  01:42:14  obrienj
// Initial revision
//
//
//-------------------------------------------------------------------
//-------------------------------------------------------------------


#include "smMatrix.H"
#include <math.h>

using namespace std;

//-------------------------------------------------------------------

inline static
istream &eatChar(char c,istream &buf) {
  char r;
  buf >> r;
  if (r!=c) {
    buf.clear(buf.rdstate() | ios::failbit);
  }
  return buf;
}

//-------------------------------------------------------------------
//-------------------------------------------------------------------
//-------------------------------------------------------------------
//-------------------------------------------------------------------
//  Implementation for  SmMatrix3x3

istream &operator>>(istream &strm, SmMatrix3x3 &m) {
  ios::fmtflags orgFlags = strm.setf(ios::skipws);
  eatChar('[',strm);

  register int r,c;

  for(r=0;r<3;r++) {
    eatChar('[',strm);
    for(c=0;c<3;c++) {
      strm >> m(r,c);
      if(c!=2) eatChar(',',strm);
    }
    eatChar(']',strm);
  }

  eatChar(']',strm);
  strm.flags(orgFlags);
  return strm;
}
  
ostream &operator<<(ostream &strm,const SmMatrix3x3 &m) {
  strm << "[";
  register int r,c;

  for(r=0;r<3;r++) {
    strm << "[ ";
    for(c=0;c<3;c++) {
      strm << m(r,c);
      if(c!=2) strm << ",";
    }
    strm << " ]";
    if(r!=2) strm << "\n ";
  }
  
  strm << " ]\n";

  return strm;
}


//-------------------------------------------------------------------

SmMatrix3x3 &SmMatrix3x3::inplaceMultPre (const SmMatrix3x3 &that) {
  SmVector3 tmp;

  register int i;

  for(i=0;i<3;i++) {

    tmp[0] = ( that.data[0][0] * data[0][i] +
	       that.data[0][1] * data[1][i] +
	       that.data[0][2] * data[2][i] );

    tmp[1] = ( that.data[1][0] * data[0][i] +
	       that.data[1][1] * data[1][i] +
	       that.data[1][2] * data[2][i] );

    tmp[2] = ( that.data[2][0] * data[0][i] +
	       that.data[2][1] * data[1][i] +
	       that.data[2][2] * data[2][i] );

    data[0][i] = tmp[0];
    data[1][i] = tmp[1];
    data[2][i] = tmp[2];
  }

  return (*this);
}


//-------------------------------------------------------------------

SmMatrix3x3 &SmMatrix3x3::inplaceMultPost(const SmMatrix3x3 &that) {
  SmVector3 tmp;

  register int i;
  
  for(i=0;i<3;i++) {
    
    tmp[0] = ( data[i][0] * that.data[0][0] +
	       data[i][1] * that.data[1][0] +
	       data[i][2] * that.data[2][0] );

    tmp[1] = ( data[i][0] * that.data[0][1] +
	       data[i][1] * that.data[1][1] +
	       data[i][2] * that.data[2][1] );

    tmp[2] = ( data[i][0] * that.data[0][2] +
	       data[i][1] * that.data[1][2] +
	       data[i][2] * that.data[2][2] );

    data[i][0] = tmp[0];
    data[i][1] = tmp[1];
    data[i][2] = tmp[2];
  }

  return (*this);
}

//-------------------------------------------------------------------

#define SWAP(a,b) { tmp = (a) ; (a) = (b) ; (b) = tmp; }

static inline double smFabs(double x) { 
  return (x>0)?(x):(-x); 
}

static void rowswap(SmMatrix3x3 &m, int i, int j) {
  register double tmp;
  SWAP(m(i,0),m(j,0));
  SWAP(m(i,1),m(j,1));
  SWAP(m(i,2),m(j,2));
}


SmMatrix3x3 inverse(const SmMatrix3x3 &a) {
  SmMatrix3x3 _a(a);
  SmMatrix3x3 _b   ; _b.setIdentity();
  int i,j,i1;
  for(j=0;j<3;j++) {
    i1 = j;
    for(i=j+1;i<3;i++)
      if (smFabs(_a(i,j)) > smFabs(_a(i1,j)))
	i1 = i;
    rowswap(_a,i1,j);
    rowswap(_b,i1,j);
    if (((_a(j,j) >= 0.0) && (_a(j,j) <  1e-6)) ||
	((_a(j,j) <= 0.0) && (_a(j,j) > -1e-6)) ){
      cerr << "Inverse of singular matrix\n" << flush;
      abort();
    }
    register double div = 1.0/_a(j,j);
    _b(j,0) *= div;
    _b(j,1) *= div;
    _b(j,2) *= div;
    _a(j,0) *= div;
    _a(j,1) *= div;
    _a(j,2) *= div;
    for(i=0;i<3;i++) {
      if (i != j) {
	register double tmp = _a(i,j);
	_b(i,0) -= tmp*_b(j,0);
	_b(i,1) -= tmp*_b(j,1);
	_b(i,2) -= tmp*_b(j,2);
	_a(i,0) -= tmp*_a(j,0);
	_a(i,1) -= tmp*_a(j,1);
	_a(i,2) -= tmp*_a(j,2);
      }
    }
  }
#ifdef DEBUG
  SmMatrix3x3 test = a*_b;
  for(i=0;i<3;i++) {
    for(j=0;j<3;j++) {
      if (smFabs(test(i,j)-((i==j)?(1):(0))) > 1e-6) {
	cerr << "There is a bug in the invese routine for SmMatrix3x3.\n";
	abort();
      }
    }
  }
#endif
  return _b;
}
	

//-------------------------------------------------------------------
//-------------------------------------------------------------------
//-------------------------------------------------------------------
//-------------------------------------------------------------------

// Lots of support code for SmSymetricEigenDecomp (SmMatrix3x3)
// This code is also used by the SVD routines.

static const double EigenEps = 1e-8;

#define EIGEN_ROTATE(a,i,j,k,l) {		\
  g=a(i,j);					\
  h=a(k,l);					\
  a(i,j)=g-s*(h+g*tau); 			\
  a(k,l)=h+s*(g-h*tau);				\
}

static inline double eigenFabs(double x) { 
  return (x>0)?(x):(-x); 
}

static inline double eigenMax(double a, double b) { 
  return (a>b)?(a):(b); 
}

static inline double eigenEpsEq(double a, double b) {
  return (eigenFabs((a)-(b)) < EigenEps);
}
  
static inline double eigenEpsEqZero(double a) {
  return (eigenFabs(a) < EigenEps);
}
  
static inline double eigenSqr(double x) { 
  return x * x; 
}

static inline double eigenPythag(double a, double b) {
  double absa,absb;
  absa=eigenFabs(a);
  absb=eigenFabs(b);
  if (absa > absb) return absa*sqrt(1.0+eigenSqr(absb/absa));
  else return (absb == 0.0 ? 0.0 : absb*sqrt(1.0+eigenSqr(absa/absb)));
}

static inline double eigenSign(double a, double b) {
  return ((b) >= 0.0 ? eigenFabs(a) : -eigenFabs(a));
}


static void eigenTred2( SmMatrix3x3 &a, SmVector3 &d, SmVector3 &e) {
  // Adapted from the Numerical Recipies in C routine "tred2"

  double scale,hh,h,g,f;

  scale = eigenFabs(a(2,0)) + eigenFabs(a(2,1));
  
  if (eigenEpsEqZero(scale)) {
    e[2]=a(2,1);
    h = 0.0;
  } else {
    a(2,0) /= scale;
    a(2,1) /= scale;
    h = a(2,0)*a(2,0) + a(2,1)*a(2,1);
    f = a(2,1);
    g = (f >= 0.0 ? -sqrt(h) : sqrt(h));
    e[2] = scale*g;
    h -= f*g;
    a(2,1) = f-g;

    a(0,2) = a(2,0)/h;
    g = a(0,0)*a(2,0) + a(1,0)*a(2,1);
    e[0]=g/h;
    f = e[0]*a(2,0);
    
    a(1,2) = a(2,1)/h;
    g = a(1,0)*a(2,0) + a(1,1)*a(2,1);
    e[1] = g/h;
    f += e[1]*a(2,1);

    hh=f/(h+h);

    f = a(2,0);
    e[0]=g=e[0]-hh*f;
    a(0,0) -= (f*e[0]+g*a(2,0));
    
    f = a(2,1);
    e[1] = g = e[1]-hh*f;
    a(1,0) -= (f*e[0]+g*a(2,0));
    a(1,1) -= (f*e[1]+g*a(2,1));
  }
  d[2]=h;

  e[1]=a(1,0);
  e[0]=0.0;

  d[0]=a(0,0);
  a(0,0)=1.0;
  
  d[1]=a(1,1);
  a(1,1)=1.0;
  a(0,1)=a(1,0)=0.0;

  if (d[2]) {
    g = a(2,0)*a(0,0);
    a(0,0) -= g*a(0,2);
    a(1,0) -= g*a(1,2);
    
    g = a(2,1);
    a(0,1) -= g*a(0,2);
    a(1,1) -= g*a(1,2);
  }
  d[2]=a(2,2);
  a(2,2)=1.0;
  a(0,2)=a(2,0)=0.0;
  a(1,2)=a(2,1)=0.0;
}



static void eigenTqli(SmVector3 &d, SmVector3 &e, SmMatrix3x3 &z) {
  // Adapted from the Numerical Recipies in C routine "tqli"

  int m,l,iter,i,k;
  double s,r,p,g,f,dd,c,b;

  for (i=1;i<3;i++) e[i-1]=e[i];
  e[2]=0.0;

  for (l=0;l<3;l++) {

    iter=0;
    do {
      for (m=l;m<2;m++) {
	dd= eigenFabs(d[m])+eigenFabs(d[m+1]);
	if (eigenEpsEq((eigenFabs(e[m])+dd),dd)) break;
      }
      if (m != l) {
	if (iter++ >= 30) {
	  cerr << "SmSymetricEigenDecomp:: Too many iterations in eigenTqli.\n";
	  return;
	}
	g=(d[l+1]-d[l])/(2.0*e[l]);
	r=eigenPythag(g,1.0);
	g=d[m]-d[l]+e[l]/(g+eigenSign(r,g));
	s=c=1.0;
	p=0.0;
	for (i=m-1;i>=l;i--) {
	  f=s*e[i];
	  b=c*e[i];
	  e[i+1]=(r=eigenPythag(f,g));
	  if (eigenEpsEqZero(r)) {
	    d[i+1] -= p;
	    e[m]=0.0;
	    break;
	  }
	  s=f/r;
	  c=g/r;
	  g=d[i+1]-p;
	  r=(d[i]-g)*s+2.0*c*b;
	  d[i+1]=g+(p=s*r);
	  g=c*r-b;
	  for (k=0;k<3;k++) {
	    f=z(k,i+1);
	    z(k,i+1)=s*z(k,i)+c*f;
	    z(k,i)=c*z(k,i)-s*f;
	  }
	}
	if (eigenEpsEqZero(r) && i >= l) continue;
	d[l] -= p;
	e[l]=g;
	e[m]=0.0;
      }
    } while (m != l);
  }
}

void SmSymetricEigenDecomp(const SmMatrix3x3 &a,
			   /* */ SmVector3   &vals,
			   /* */ SmMatrix3x3 &vecs) {
  SmVector3 e;
  SmMatrix3x3 q(a);

#ifdef DEBUG
  if (!(eigenEpsEq( q(0,1),q(1,0) ) &&
	eigenEpsEq( q(2,1),q(1,2) ) &&
	eigenEpsEq( q(2,0),q(0,2) ) )) {
    cerr << "SmSymetricEigenDecomp called with non-symetric matrix.\n";
    abort();
  }
#endif

  vals = 0.0;
  eigenTred2(q,vals,e);
  vecs.setIdentity();  
  eigenTqli(vals,e,vecs);
  vecs.inplaceMultPre(transpose(q));
}


//-------------------------------------------------------------------
//-------------------------------------------------------------------

static 
void SmSVDecompInternal(SmMatrix3x3 &a, SmVector3 &w, SmMatrix3x3 &v) {

  // Adapted from the Numerical Recipies in C routine "dsvdcmp"

  int flag,i,its,j,jj,k,l,nm;
  double anorm,c,f,g,h,s,scale,x,y,z;

  SmVector3 rv1;
  g=scale=anorm=0.0;
  for (i=0;i<3;i++) {
    l=i+1;
    rv1[i]=scale*g;
    g=s=scale=0.0;
    if (i < 3) {
      for (k=i;k<3;k++) scale += eigenFabs(a(k,i));
      if (scale) {
	for (k=i;k<3;k++) {
	  a(k,i) /= scale;
	  s += a(k,i)*a(k,i);
	}
	f=a(i,i);
	g = -eigenSign(sqrt(s),f);
	h=f*g-s;
	a(i,i)=f-g;
	for (j=l;j<3;j++) {
	  for (s=0.0,k=i;k<3;k++) s += a(k,i)*a(k,j);
	  f=s/h;
	  for (k=i;k<3;k++) a(k,j) += f*a(k,i);
	}
	for (k=i;k<3;k++) a(k,i) *= scale;
      }
    }
    w[i]=scale *g;
    g=s=scale=0.0;
    if (i < 3 && i != 2 ) {
      for (k=l;k<3;k++) scale += eigenFabs(a(i,k));
      if (scale) {
	for (k=l;k<3;k++) {
	  a(i,k) /= scale;
	  s += a(i,k)*a(i,k);
	}
	f=a(i,l);
	g = -eigenSign(sqrt(s),f);
	h=f*g-s;
	a(i,l)=f-g;
	for (k=l;k<3;k++) rv1[k]=a(i,k)/h;
	for (j=l;j<3;j++) {
	  for (s=0.0,k=l;k<3;k++) s += a(j,k)*a(i,k);
	  for (k=l;k<3;k++) a(j,k) += s*rv1[k];
	}
	for (k=l;k<3;k++) a(i,k) *= scale;
      }
    }
    anorm=eigenMax(anorm,(eigenFabs(w[i])+eigenFabs(rv1[i])));
  }


  for (i=2;i>=0;i--) {
    if (i < 2) {
      if (g) {
	for (j=l;j<3;j++) v(j,i)=(a(i,j)/a(i,l))/g;
	for (j=l;j<3;j++) {
	  for (s=0.0,k=l;k<3;k++) s += a(i,k)*v(k,j);
	  for (k=l;k<3;k++) v(k,j) += s*v(k,i);
	}
      }
      for (j=l;j<3;j++) v(i,j)=v(j,i)=0.0;
    }
    v(i,i)=1.0;
    g=rv1[i];
    l=i;
  }
  for (i=2;i>=0;i--) {
    l=i+1;
    g=w[i];
    for (j=l;j<3;j++) a(i,j)=0.0;
    if (g) {
      g=1.0/g;
      for (j=l;j<3;j++) {
	for (s=0.0,k=l;k<3;k++) s += a(k,i)*a(k,j);
	f=(s/a(i,i))*g;
	for (k=i;k<3;k++) a(k,j) += f*a(k,i);
      }
      for (j=i;j<3;j++) a(j,i) *= g;
    } else for (j=i;j<3;j++) a(j,i)=0.0;
    ++a(i,i);
  }

  for (k=2;k>=0;k--) {
    for (its=1;its<=30;its++) {
      flag=1;
      for (l=k;l>=0;l--) {
	nm=l-1;
	if ((double)(eigenFabs(rv1[l])+anorm) == anorm) {
	  flag=0;
	  break;
	}
	if ((double)(eigenFabs(w[nm])+anorm) == anorm) break;
      }

      if (flag) {
	c=0.0;
	s=1.0;
	for (i=l;i<=k;i++) {
	  f=s*rv1[i];
	  rv1[i]=c*rv1[i];
	  if ((double)(eigenFabs(f)+anorm) == anorm) break;
	  g=w[i];
	  h=eigenPythag(f,g);
	  w[i]=h;
	  h=1.0/h;
	  c=g*h;
	  s = -f*h;
	  for (j=0;j<3;j++) {
	    y=a(j,nm);
	    z=a(j,i);
	    a(j,nm)=y*c+z*s;
	    a(j,i)=z*c-y*s;
	  }
	}
      }
      z=w[k];
      if (l == k) {
	if (z < 0.0) {
	  w[k] = -z;
	  for (j=0;j<3;j++) v(j,k) = -v(j,k);
	}
	break;
      }
      if (its == 30) {
	cerr << "SmSVDecomp:: Too many iterations in SmSVDecompInternal.\n";
	return;
      }
      x=w[l];
      nm=k-1;
      y=w[nm];
      g=rv1[nm];
      h=rv1[k];
      f=((y-z)*(y+z)+(g-h)*(g+h))/(2.0*h*y);
      g=eigenPythag(f,1.0);
      f=((x-z)*(x+z)+h*((y/(f+eigenSign(g,f)))-h))/x;
      c=s=1.0;

      for (j=l;j<=nm;j++) {
	i=j+1;
	g=rv1[i];
	y=w[i];
	h=s*g;
	g=c*g;
	z=eigenPythag(f,h);
	rv1[j]=z;
	c=f/z;
	s=h/z;
	f=x*c+g*s;
	g = g*c-x*s;
	h=y*s;
	y *= c;
	for (jj=0;jj<3;jj++) {
	  x=v(jj,j);
	  z=v(jj,i);
	  v(jj,j)=x*c+z*s;
	  v(jj,i)=z*c-x*s;
	}
	z=eigenPythag(f,h);
	w[j]=z;
	if (z) {
	  z=1.0/z;
	  c=f*z;
	  s=h*z;
	}
	f=c*g+s*y;
	x=c*y-s*g;
	for (jj=0;jj<3;jj++) {
	  y=a(jj,j);
	  z=a(jj,i);
	  a(jj,j)=y*c+z*s;
	  a(jj,i)=z*c-y*s;
	}
      }
      rv1[l]=0.0;
      rv1[k]=f;
      w[k]=x;
    }
  }
}




void SmSVDecomp(const SmMatrix3x3 &a,
		/* */ SmMatrix3x3 &u,
		/* */ SmVector3   &w,
		/* */ SmMatrix3x3 &v) {
  u = a;
  SmSVDecompInternal(u,w,v);
}

 
//-------------------------------------------------------------------
//-------------------------------------------------------------------
//-------------------------------------------------------------------
//-------------------------------------------------------------------
//  Implementation for  SmMatrix2x2

istream &operator>>(istream &strm, SmMatrix2x2 &m) {
  ios::fmtflags orgFlags = strm.setf(ios::skipws);
  eatChar('[',strm);
  
  register int r,c;
  
  for(r=0;r<2;r++) {
    eatChar('[',strm);
    for(c=0;c<2;c++) {
      strm >> m(r,c);
      if(c!=1) eatChar(',',strm);
    }
    eatChar(']',strm);
  }

  eatChar(']',strm);
  strm.flags(orgFlags);
  return strm;
}
  
ostream &operator<<(ostream &strm,const SmMatrix2x2 &m) {
  strm << "[";
  register int r,c;
  
  for(r=0;r<2;r++) {
    strm << "[ ";
    for(c=0;c<2;c++) {
      strm << m(r,c);
      if(c!=1) strm << ",";
    }
    strm << " ]";
    if(r!=1) strm << "\n ";
  }
  strm << " ]\n";
  return strm;
}

//-------------------------------------------------------------------
  
SmMatrix2x2 &SmMatrix2x2::inplaceMultPre (const SmMatrix2x2 &that) {
  SmVector2 tmp;
  register int i;

  for(i=0;i<2;i++) {
    tmp[0] = ( that.data[0][0] * data[0][i] +
	       that.data[0][1] * data[1][i] );
    tmp[1] = ( that.data[1][0] * data[0][i] +
	       that.data[1][1] * data[1][i] );
    data[0][i] = tmp[0];
    data[1][i] = tmp[1];
  }
  return (*this);
}


//-------------------------------------------------------------------

SmMatrix2x2 &SmMatrix2x2::inplaceMultPost(const SmMatrix2x2 &that) {
  SmVector2 tmp;
  register int i;
  for(i=0;i<2;i++) {
    tmp[0] = ( data[i][0] * that.data[0][0] +
	       data[i][1] * that.data[1][0] );
    tmp[1] = ( data[i][0] * that.data[0][1] +
	       data[i][1] * that.data[1][1] );
    data[i][0] = tmp[0];
    data[i][1] = tmp[1];
  }
  return (*this);
}

  
//-------------------------------------------------------------------

static void rowswap(SmMatrix2x2 &m, int i, int j) {
  register double tmp;
  SWAP(m(i,0),m(j,0));
  SWAP(m(i,1),m(j,1));
}


SmMatrix2x2 inverse(const SmMatrix2x2 &a) {
  SmMatrix2x2 _a(a);
  SmMatrix2x2 _b   ; _b.setIdentity();
  int i,j,i1;
  for(j=0;j<2;j++) {
    i1 = j;
    for(i=j+1;i<2;i++)
      if (smFabs(_a(i,j)) > smFabs(_a(i1,j)))
	i1 = i;
    rowswap(_a,i1,j);
    rowswap(_b,i1,j);
    if (_a(j,j) == 0.0) {
      cerr << "Inverse of singular matrix\n" << flush;
      abort();
    }
    register double div = 1.0/_a(j,j);
    _b(j,0) *= div;
    _b(j,1) *= div;
    _a(j,0) *= div;
    _a(j,1) *= div;
    for(i=0;i<2;i++) {
      if (i != j) {
	register double tmp = _a(i,j);
	_b(i,0) -= tmp*_b(j,0);
	_b(i,1) -= tmp*_b(j,1);
	_a(i,0) -= tmp*_a(j,0);
	_a(i,1) -= tmp*_a(j,1);
      }
    }
  }
#ifdef DEBUG
  SmMatrix2x2 test = a*_b;
  for(i=0;i<2;i++) {
    for(j=0;j<2;j++) {
      if (smFabs(test(i,j)-((i==j)?(1):(0))) > 1e-10) {
	cerr << "There is a bug in the invese routine for SmMatrix2x2.\n";
	abort();
      }
    }
  }
#endif
  return _b;
}
	
//-------------------------------------------------------------------
//-------------------------------------------------------------------

void SmSymetricEigenDecomp(const SmMatrix2x2 &a,
			   /* */ SmVector2   &vals,
			   /* */ SmMatrix2x2 &vecs) {
#ifdef DEBUG
    if (a(1,0) != a(0,1)) {
      cerr << "SmSymetricEigenDecomp :: Routine called with unsymetric matrix.\n" << flush;
      abort();
    }
    if (isnan(a(1,0))) {
      cerr << "SmSymetricEigenDecomp :: Routine called with NaN matrix.\n" << flush;
      abort();
    }
#endif

  double ct = sqrt(4*eigenSqr(a(0,1)) + eigenSqr(a(0,0) - a(1,1)));
                                     
  double val0 = 0.5*(a(0,0) + a(1,1) - ct);
  double val1 = 0.5*(a(0,0) + a(1,1) + ct);

  double d = (a(0,0) >= a(1,1)) ?
    sqrt(eigenSqr(a(0,1)) + 0.25*eigenSqr( a(0,0) - a(1,1) + ct )) :
    sqrt(eigenSqr(a(0,1)) + 0.25*eigenSqr(-a(0,0) + a(1,1) + ct )) ;

  double p;
  double q;

  if ((d+1.0) == 1.0) {
    p = 0.0;
    q = 0.0;
  }else{
    d = 1.0 / d;
    if (a(0,0) >= a(1,1)) {
      p = d * a(0,1);
      q = d * 0.5 * (-a(0,0) + a(1,1) - ct );
    }else{
      q = d * a(0,1);
      p = d * 0.5 * ( a(0,0) - a(1,1) - ct );
    }
  }
      
  vals[0] = val0;
  vals[1] = val1;

  vecs(0,0) = p;
  vecs(0,1) = q;

  vecs(1,0) = -q;
  vecs(1,1) =  p;
}





//-------------------------------------------------------------------
//-------------------------------------------------------------------
//-------------------------------------------------------------------
//-------------------------------------------------------------------

